# Copyright (C) 2017-2020 RAW Labs SA
import base64
import configparser
import itertools
import json
import logging.config
import os
import pathlib
import re
import requests
import six
import ssl
import tempfile
import time
import urllib3
import websocket
from collections import OrderedDict
from datetime import datetime, timedelta
from dateutil.relativedelta import relativedelta
from io import RawIOBase, BufferedReader, TextIOWrapper
from six.moves.urllib.parse import urlparse

try:
    from urllib.parse import urljoin
    from urllib.request import pathname2url
    from urllib.parse import quote_plus
except ImportError:
    from urlparse import urljoin
    from urllib import pathname2url
    from urllib import quote_plus

INI_FILE = 'raw.ini'

ENV_RAW_CLIENT_CONFIG_FILE = 'RAW_CLIENT_CONFIG_FILE'
ENV_RAW_ACCESS_TOKEN = 'RAW_ACCESS_TOKEN'
ENV_RAW_ACCESS_TOKEN_FILE = 'RAW_ACCESS_TOKEN_FILE'

_RAW_OUTPUT_FORMAT = 'output-format'
_RAW_WINDOWS_LINE_ENDING = 'windows-line-ending'

QUERY_TEXT_FORMATS = ['json', 'hjson', 'raw', 'csv']
QUERY_FORMATS = QUERY_TEXT_FORMATS + ['sqlite', 'parquet-avro', 'parquet-raw']
QUERY_LOG_LEVELS = ['none', 'warn', 'info', 'debug', 'trace']

# _COMMON_HEADERS = {'Accept-Encoding': 'identity'}  # No gzip, using plain text for wireshark capture
_COMMON_HEADERS = {}

# Create logger
try:
    # Try to load logging.ini
    # TODO (ns): Improve loading of logging.ini file. Use a similar system as logback.
    logging.config.fileConfig('logging.ini')
    logger = logging.getLogger(__name__)
except Exception as e:
    logger = logging.getLogger(__name__)
    ch = logging.StreamHandler()
    ch.setLevel(logging.DEBUG)
    logger.addHandler(ch)

class ServerConfig:
    def __init__(self, config_parser):
        self.executor_url = config_parser.get('raw', 'executor_url', fallback=None)
        if self.executor_url and self.executor_url.endswith('/'):
            self.executor_url = self.executor_url[:-1]

        self.creds_url = config_parser.get('raw', 'creds_url', fallback=None)
        if self.creds_url and self.creds_url.endswith('/'):
            self.creds_url = self.creds_url[:-1]

        self.kafka_url = config_parser.get('raw', 'kafka_url', fallback=None)
        if self.kafka_url and self.kafka_url.endswith('/'):
            self.kafka_url = self.kafka_url[:-1]

        self.storage_url = config_parser.get('raw', 'storage_url', fallback=None)
        if self.storage_url and self.storage_url.endswith('/'):
            self.storage_url = self.storage_url[:-1]

        disable_tls_cert_verification = config_parser.getboolean('raw', 'disable_tls_cert_verification', fallback=False)
        self.verify_tls_cert = not disable_tls_cert_verification if disable_tls_cert_verification else None

        self.http_proxy = config_parser.get('raw', 'http_proxy', fallback=None)
        self.https_proxy = config_parser.get('raw', 'https_proxy', fallback=None)
        self.query_response_transfer = config_parser.get('raw', 'query_response_transfer', fallback='websockets')
        self.query_log_level = config_parser.get('raw', 'query_log_level', fallback=None)
        if self.query_log_level is None:
            if self.query_response_transfer == 'websockets':
                self.query_log_level = 'info'
            else:
                self.query_log_level = 'none'
        self.connect_timeout = config_parser.getfloat('raw', 'connect_timeout', fallback=15.0)
        self.read_timeout = config_parser.getfloat('raw', 'read_timeout', fallback=None)


        self.auth = config_parser.get('raw', 'auth', fallback=None)
        self.auth_basic_access_token = config_parser.get('auth_basic', 'access_token', fallback=None)
        self.oauth2_provider = config_parser.get('auth_oauth2', 'provider', fallback=None)
        self.oauth2_grant_type = config_parser.get('auth_oauth2', 'grant_type', fallback=None)
        self.oauth2_domain = config_parser.get('auth_oauth2', 'domain', fallback=None)
        self.oauth2_audience = config_parser.get('auth_oauth2', 'audience', fallback=None)
        self.oauth2_client_id = config_parser.get('auth_oauth2', 'client_id', fallback=None)
        self.oauth2_client_secret = config_parser.get('auth_oauth2', 'client_secret', fallback=None)
        # Directory where to save the OAuth2 tokens
        self.oauth2_tokens_directory = config_parser.get('auth_oauth2', 'token_location',
                                                         fallback=str(pathlib.Path.home().joinpath('.rawcli')))
    def __str__(self):
        return '{} {} {} {} {} {} {}'.format(self.executor_url, self.creds_url, self.kafka_url, self.auth,
                                             self.oauth2_domain, self.oauth2_client_id, self.oauth2_audience)


class HttpInputStream(RawIOBase):
    """Creates an input stream from out /query rest service"""

    def __init__(self, response):
        self.query_id = response.headers['X-RAW-id']
        self.pos = 0
        self.has_more = True
        self.data = None
        self.response = response
        self.chunks = response.iter_content(chunk_size=None, decode_unicode=False)

    def read(self, size=-1):
        if not self.has_more:
            return b''
        if not self.data:
            try:
                self.data = next(self.chunks)
                self.pos = 0
            except StopIteration:
                self.has_more = False
                return b''

        if size < 0:
            out = self.readall()
        elif size >= (len(self.data) - self.pos):
            out = self.data[self.pos:]
            self.data = None
        else:
            out = self.data[self.pos:self.pos + size]
            self.pos = self.pos + size
        return out

    # TODO: check why this method has to be implemented instead of using a compile c one
    def readinto(self, buffer):
        pos = 0
        while True:
            bytes = self.read(len(buffer) - pos)
            buffer[pos:pos + len(bytes)] = bytes
            pos += len(bytes)
            if pos >= len(buffer) or not len(bytes):
                return pos

    def writable(self):
        return False

    def readable(self):
        return True

    def close(self):
        self.has_more = False
        self.data = None
        self.chunks.close()
        try:
            self.chunks.close()
        finally:
            pass
        try:
            self.response.close()
        finally:
            pass


def _parse_proxy(proxy):
    proxy = urlparse(proxy)
    proxy_hostname = proxy.hostname
    proxy_auth = (proxy.username, proxy.password) if proxy.username else None
    proxy_port = proxy.port
    return proxy_hostname, proxy_port, proxy_auth

class WebSocketInputStream(RawIOBase):
    """Creates an input stream out of the WebSocket protocol (ws/query).
    """

    def __init__(self, url, token, query, options, log_level, msg_callback, as_sql, config):
        self.pos = 0
        self.data = None
        self.has_more = True
        self.msg_callback = msg_callback
        self.as_sql = as_sql
        if not url:
            raise RawException('Executor service endpoint is not configured')

        # Convert the http[s]:// url into a websockets address
        ws_address = "ws" + url[4:]
        if config.verify_tls_cert:
            sslopt = {}
        else:
            sslopt = {"cert_reqs": ssl.CERT_NONE}

        if ws_address.startswith('ws://') and config.http_proxy:
            proxy_host, proxy_port, proxy_auth = _parse_proxy(config.http_proxy)
        elif ws_address.startswith('wss://') and config.https_proxy:
            proxy_host, proxy_port, proxy_auth = _parse_proxy(config.https_proxy)
        else:
            proxy_host = None
            proxy_port = 0
            proxy_auth = None

        self.start_time = time.time()
        self.ws = None
        try:
            self.ws = websocket.create_connection('{}/1/public/ws/query'.format(ws_address), timeout=config.connect_timeout,
                sslopt=sslopt, http_proxy_host=proxy_host, http_proxy_port=proxy_port,
                http_proxy_auth=proxy_auth)
            self.ws.settimeout(config.read_timeout)
            if log_level == 'none': request_log_level = None
            else: request_log_level = log_level
            query_request = dict(accessToken=token, query=query, options=options, logLevel=request_log_level, asSql=as_sql)
            self.ws.send(json.dumps(query_request))
            while True:
                opcode, frame = self.ws.recv_data_frame()
                if opcode is not websocket.ABNF.OPCODE_TEXT:
                    if opcode == websocket.ABNF.OPCODE_CLOSE:
                        if frame.data and len(frame.data) >= 2:
                            code = 256 * six.byte2int(frame.data[0:1]) + six.byte2int(frame.data[1:2])
                            reason = frame.data[2:].decode('utf-8')
                        else:
                            code = "<not set>"
                            reason = "<not set>"
                        raise RawException('Socket closed unexpectedly.\nCode: {}\nReason: {}'.format(code, reason))
                    else:
                        raise RawException('Unexpected response.\nOpcode: {}\nFrame: {}'.format(opcode, frame))
                result_msg = self.__handle_text_message(frame)
                if result_msg['type'] == 'unauthorized':
                    raise Unauthorized(result_msg['error'])
                if result_msg['type'] == 'invalidSession':
                    raise RawException(result_msg['error'])
                elif result_msg['type'] == 'invalidQuery':
                    raise QueryValidationError(result_msg['errors'])
                elif result_msg['type'] == 'querySuccess':
                    self.query_id = result_msg['id']
                    return
                elif result_msg['type'] not in ['userMessage']:
                    raise RawException('Unexpected response.\nUnexpected control message: {}'.format(result_msg))
        except OSError as ex:
            self.close()
            raise RawException(
                'Service unavailable. Connection error: is server running?\nResponse body: {}'.format(ex))
        except websocket.WebSocketBadStatusException as ex:
            self.close()
            if ex.status_code == 503:
                raise RawException(
                    'Service unavailable. Please try again later.\nResponse status: {}\nResponse body: {}'.format(
                        ex.status_code, str(ex)))
            else:
                raise RawException(
                    'Unexpected response. Response status: {}\nResponse body: {}'.format(ex.status_code, str(ex)))
        except websocket.WebSocketConnectionClosedException:
            self.close()
            raise RawException("Connection closed unexpectedly")

        except (KeyboardInterrupt, SystemExit, Exception):
            self.close()
            raise

    def __handle_text_message(self, frame):
        control_msg = json.loads(frame.data.decode('utf-8'))
        if control_msg['type'] == 'userMessage':
            if control_msg['level'] == 'error':
                raise QueryExecutionError(control_msg['message'])
            elif self.msg_callback is not None:
                self.msg_callback(control_msg['level'],
                                  '[{0:.3f}] {1}'.format(time.time() - self.start_time, control_msg['message']))
        elif control_msg['type'] not in ['unauthorized', 'invalidSession', 'invalidQuery', 'querySuccess', 'queryStreamingEnd']:
            raise RawException('Unexpected response.\nUnknown control message: {}'.format(control_msg))
        return control_msg

    def read(self, size=-1):
        if not self.has_more:
            return b''
        while not self.data:
            opcode, frame = self.ws.recv_data_frame()
            if opcode == websocket.ABNF.OPCODE_BINARY:
                self.data = frame.data
                self.pos = 0
            elif opcode == websocket.ABNF.OPCODE_TEXT:
                # Control message: error, warning or end of stream
                control_msg = self.__handle_text_message(frame)
                if control_msg['type'] == 'queryStreamingEnd':
                    self.has_more = False
                    return b''
                elif control_msg['type'] not in ['userMessage']:
                    raise RawException(
                        'Unexpected response.\nUnexpected control message during read: {}'.format(control_msg))

        if size < 0:
            out = self.readall()
        elif size >= (len(self.data) - self.pos):
            out = self.data[self.pos:]
            self.data = None
        else:
            out = self.data[self.pos:self.pos + size]
            self.pos = self.pos + size
        return out

    def readinto(self, buffer):
        pos = 0
        while True:
            bytes = self.read(len(buffer) - pos)
            buffer[pos:pos + len(bytes)] = bytes
            pos += len(bytes)
            if pos >= len(buffer) or not len(bytes):
                return pos

    def writable(self):
        return False

    def readable(self):
        return True

    def close(self):
        self.has_more = False
        self.data = None
        if self.ws:
            self.ws.close()
            self.ws = None


class RawPythonIterator:
    """Creates a iterator of Python objects from a query output stream."""

    def __init__(self, stream, tipe, np_arrays, use_relativedeltas):
        assert tipe['type'] == 'collection'
        self.__stream = stream
        self.__is_closed = False
        self.tipe = tipe
        self.np_arrays = np_arrays
        self.relativedeltas = use_relativedeltas

    def __iter__(self):
        return self

    def __next__(self):
        return self.next()

    def next(self):
        try:
            line = self.__stream.readline()
            if not line:
                raise StopIteration
            obj = json.loads(line, object_pairs_hook=OrderedDict)
            return convert_types(obj, self.tipe['inner'], np_arrays=self.np_arrays,
                                 use_relativedeltas=self.relativedeltas)
        except (KeyboardInterrupt, SystemExit, StopIteration, Exception):
            self.close()
            raise

    def type(self):
        return self.tipe

    def close(self):
        if not self.__is_closed:
            self.__stream.close()
            self.__is_closed = True


# Wrapper around TextIOWrapper that holds member with query_id
class RawTextIOWrapper(TextIOWrapper):
    def __init__(self, stream, encoding):
        super().__init__(BufferedReader(stream), encoding=encoding)
        self.query_id = stream.query_id


def find_first_existing(locations):
    for l in locations:
        p = os.path.abspath(l)
        if os.path.exists(p):
            return p
        else:
            logger.debug('File not found: %s', l)
    return None


def load_config(config_file):
    if config_file is not None:
        if not os.path.exists(config_file):
            raise ConfigurationException('Configuration file not found: {}'.format(config_file))
        locations = [config_file]
    else:
        # Search order of ini files:
        # 1. contents of RAW_CLIENT_CONFIG_FILE env variable if defined
        # 2. raw.ini in current working directory
        # 3. .raw.ini in home directory
        # In the future, maybe /etc/raw.ini
        locations = []
        if ENV_RAW_CLIENT_CONFIG_FILE in os.environ:
            locations.append(os.environ[ENV_RAW_CLIENT_CONFIG_FILE])
        locations.append(INI_FILE)
        locations.append(os.path.join(os.path.expanduser('~'), INI_FILE))
        locations.append(os.path.join(os.path.expanduser('~'), '.{}'.format(INI_FILE)))

    ini_file = find_first_existing(locations)
    logger.debug('Loading configuration file: %s', ini_file)
    if ini_file is None:
        raise ConfigurationException('No configuration file found. Searched: [{}]'.format(', '.join(locations)))

    config_parser = configparser.ConfigParser()
    try:
        config_parser.read(ini_file)
        server_config = ServerConfig(config_parser)
        logger.debug('Configuration %s', server_config)
        return server_config
    except configparser.NoSectionError as e:
        raise ConfigurationException('Could not find configuration file ({}): {}'.format(ini_file, e))
    except configparser.Error as e:
        raise ConfigurationException('Error in config file ({}): {}'.format(ini_file, e))


def load_oauth2_access_token(token_file):
    """The file must contain aJSON object as returned by the token endpoint of an OAuth2 implementation, with fields
    "access_token" and/or "id_token"
    """
    if not os.path.exists(token_file):
        raise RawException('File with access token not found: {}'.format(token_file))
    logger.debug("Loading token from file: {}".format(token_file))
    with open(token_file, 'r') as f:
        file_contents = f.read()
    try:
        return json.loads(file_contents)
    except json.decoder.JSONDecodeError as ex:
        raise RawException("Could not load file: {}. Error: {}".format(token_file, ex))


def save_oauth2_access_token(token_file, token):
    # expires_in = response['expires_in']
    # from datetime import datetime, timedelta
    # expires_by = datetime.now() + timedelta(seconds=expires_in)
    # print('Configuration for using the access token (expires by: {}): '.format(
    #     expires_by.strftime('%Y-%m-%d %H:%M:%S')))
    logger.debug('Writing access tokens to: {}'.format(token_file))
    directory = os.path.dirname(token_file)
    logger.debug('dir name: {}'.format(token_file))
    if not os.path.exists(directory):
        os.makedirs(directory)
    # Writing JSON data
    with open(token_file, 'w') as f:
        json.dump(token, f, indent=2)


def get_token_file_location(config):
    if config.oauth2_grant_type == 'authorization_code':
        prefix = 'ac'
    elif config.oauth2_grant_type == 'client_credentials':
        prefix = 'cc'
    else:
        raise RawException('Invalid grant type or grant type does not use access tokens: {}'.format(config.grant_type))
    token_file = os.path.join(config.oauth2_tokens_directory, '{}-{}.json'.format(prefix, config.oauth2_client_id))
    return token_file


def new_raw_client(config_file=None,
                   executor_url=None, creds_url=None, kafka_url=None, storage_url=None,
                   access_token=None, json_token=None, access_token_file=None,
                   **kwargs):
    """
    Creates a new instance of the RAW client.
    For auth, this method checks if one of the following is defined, and uses the first one that is defined:
    - access_token -- a string with the access token (opaque or id_token/JWT)
    - json_token -- a Python map with a token, corresponding to the json returned by the /token endpoint.
        See https://developer.okta.com/docs/reference/api/oidc/#token  for a description of the expected fields
    - access_token_file -- file containing a json structure as returned by the /token endpoint.

    If none of these arguments is defined, it looks up the following env variables, also in order:
    - RAW_ACCESS_TOKEN - contains an access token (opaque or id_token/JWT)
    - RAW_ACCESS_TOKEN_FILE - Points to a file with the json returned by /token endpoint.

    If they are also not defiend, then it looks for a file with a json token structure as generated by rawcli login
    """
    logger.debug('Configuration file? %s', config_file)

    config = load_config(config_file)

    executor_url = executor_url if executor_url else config.executor_url
    creds_url = creds_url if creds_url else config.creds_url
    storage_url = storage_url if storage_url else config.storage_url
    kafka_url = kafka_url if kafka_url else config.kafka_url

    # Load access token. Priorities:
    # 1. passed as an argument
    # 2. RAW_ACCESS_TOKEN env variable
    # 3. RAW_ACCESS_TOKEN_FILE env variable pointing to a file on disk
    # 4. load from ~/.raw_oauth2_config
    if config.auth == 'oauth2':
        if access_token:
            token_provider = SimpleTokenProvider(access_token)
        elif json_token:
            token_provider = JsonTokenProvider(json_token, config)
        elif access_token_file:
            token_provider = AuthorizationCodeTokenProvider(access_token_file, config)
        elif ENV_RAW_ACCESS_TOKEN in os.environ:
            logger.debug('Using simple access token in environment variable {}'.format(ENV_RAW_ACCESS_TOKEN))
            token_provider = SimpleTokenProvider(os.environ[ENV_RAW_ACCESS_TOKEN])
        elif ENV_RAW_ACCESS_TOKEN_FILE in os.environ:
            logger.debug('Loading OAuth2 tokens from file in environment variable {}'.format(ENV_RAW_ACCESS_TOKEN_FILE))
            access_token_file = os.environ[ENV_RAW_ACCESS_TOKEN_FILE]
            token_provider = AuthorizationCodeTokenProvider(access_token_file, config)
        elif config.oauth2_grant_type == 'authorization_code':
            access_token_file = get_token_file_location(config)
            token_provider = AuthorizationCodeTokenProvider(access_token_file, config)
        elif config.oauth2_grant_type == 'client_credentials':
            access_token_file = get_token_file_location(config)
            token_provider = ClientCredentialsTokenProvider(access_token_file, config)
        else:
            raise RawException('Invalid value for configuration property "auth_oauth2.grant_type": {}'.format(config.oauth2_grant_type))
    elif config.auth == 'basic':
        token_provider = SimpleTokenProvider(config.auth_basic_access_token)
    else:
        raise RawException('Invalid value for configuration property "raw.auth": {}'.format(config.auth))

    return Raw(executor_url, creds_url, kafka_url, storage_url, token_provider, config, **kwargs)


def set_raw_client_logging_level(level):
    logger.setLevel(level)


def set_raw_client_formatter(formatter):
    logger.handlers[0].setFormatter(formatter)


class TokenProvider:
    def get_token(self):
        pass


class SimpleTokenProvider(TokenProvider):
    def __init__(self, access_token):
        self._access_token = access_token

    def get_token(self):
        return self._access_token


class JsonTokenProvider(TokenProvider):
    def __init__(self, json_token, config):
        self._json_token = json_token
        self._config = config
        self.__extract_access_token()

    def __extract_access_token(self):
        if 'id_token' in self._json_token:
            self._access_token = self._json_token['id_token']
        elif 'access_token' in self._oauth2_config:
            self._access_token = self._json_token['access_token']

    def __do_refresh_token(self):
        logger.debug("Using refresh token to obtain new access token")
        domain = self._config.oauth2_domain
        client_id = self._config.oauth2_client_id
        client_secret = self._config.oauth2_client_secret
        missing_param = []
        if client_id is None:
            missing_param.append('client_id')
        if client_secret is None:
            missing_param.append('client_secret')
        if domain is None:
            missing_param.append('domain')
        if len(missing_param) != 0:
            raise RawException('Missing parameters from configuration: {}'.format(', '.join(missing_param)))
        if 'refresh_token' not in self._json_token:
            raise RawException('No refresh token available.')

        # https://docs.cloudfoundry.org/api/uaa/version/74.23.0/index.html#refresh-token
        data = {
            'client_id': client_id,
            'client_secret': client_secret,
            'grant_type': 'refresh_token',
            'refresh_token': self._json_token['refresh_token']
        }

        # TODO: Support for refresh tokens with Auth0. This does not exist in Cloudfoundry, which is used by Swisscom
        # if provider == 'auth0':
        #     if audience is None:
        #         raise Exception('Configuration property "auth_oauth2.audience" must be specified when using client credentials with Auth0 provider')
        #     data['audience'] = audience
        headers = {'Accept': 'application/json'}
        url = 'https://{}/oauth/token'.format(domain)
        logger.debug('Obtaining OAuth2 token. URL: {}'.format(url))
        resp = requests.post(url, data=data, headers=headers, proxies=get_requests_proxy(self._config))
        if resp.status_code in [400, 401, 403, 404]:
            error = json.loads(resp.text)
            raise RawException('Could not refresh token. Run "rawcli login -f" to obtain a new token.\nError: {}'.format(error))
        elif resp.status_code != 200:
            raise RawException(
                'Unexpected response.\nResponse status: {}\nResponse body: {}'.format(resp.status_code, resp.text))
        else:
            self._json_token = resp.json()
            self.__extract_access_token()

    def get_token(self):
        if not self._access_token:
            self.__do_refresh_token()
        return self._access_token

    def refresh_token(self):
        if 'refresh_token' in self._json_token:
            self.__do_refresh_token()
            return True
        else:
            return False


class AuthorizationCodeTokenProvider(TokenProvider):
    def __init__(self, token_file, config):
        self._config = config
        self._token_file = token_file
        self.__init_from_token_file()

    def __init_from_token_file(self):
        if not os.path.exists(self._token_file):
            raise RawException('File with access token not found: {}. Run "rawcli login" to obtain a new token.'.format(self._token_file))
        self._oauth2_config = load_oauth2_access_token(self._token_file)
        if 'id_token' in self._oauth2_config:
            self._access_token = self._oauth2_config['id_token']
        elif 'access_token' in self._oauth2_config:
            self._access_token = self._oauth2_config['access_token']

    def __do_refresh_token(self):
        logger.debug("Using refresh token to obtain new access token")
        domain = self._config.oauth2_domain
        client_id = self._config.oauth2_client_id
        client_secret = self._config.oauth2_client_secret
        missing_param = []
        if client_id is None:
            missing_param.append('client_id')
        if client_secret is None:
            missing_param.append('client_secret')
        if domain is None:
            missing_param.append('domain')
        if len(missing_param) != 0:
            raise RawException('Missing parameters from configuration: {}'.format(', '.join(missing_param)))
        if 'refresh_token' not in self._oauth2_config:
            raise RawException('No refresh token available.')

        # https://docs.cloudfoundry.org/api/uaa/version/74.23.0/index.html#refresh-token
        data = {
            'client_id': client_id,
            'client_secret': client_secret,
            'grant_type': 'refresh_token',
            'refresh_token': self._oauth2_config['refresh_token']
        }

        # TODO: Support for refresh tokens with Auth0. This does not exist in Cloudfoundry, which is used by Swisscom
        # if provider == 'auth0':
        #     if audience is None:
        #         raise Exception('Configuration property "auth_oauth2.audience" must be specified when using client credentials with Auth0 provider')
        #     data['audience'] = audience
        headers = {'Accept': 'application/json'}
        url = 'https://{}/oauth/token'.format(domain)
        logger.debug('Obtaining OAuth2 token. URL: {}'.format(url))
        resp = requests.post(url, data=data, headers=headers, proxies=get_requests_proxy(self._config))
        if resp.status_code in [400, 401, 403, 404]:
            error = json.loads(resp.text)
            raise RawException('Could not refresh token. Run "rawcli login -f" to obtain a new token.\nError: {}'.format(error))
        elif resp.status_code != 200:
            raise RawException(
                'Unexpected response.\nResponse status: {}\nResponse body: {}'.format(resp.status_code, resp.text))
        else:
            save_oauth2_access_token(self._token_file, resp.json())
            self.__init_from_token_file()

    def get_token(self):
        if not self._access_token:
            self.__do_refresh_token()
        return self._access_token

    def refresh_token(self):
        if 'refresh_token' in self._oauth2_config:
            self.__do_refresh_token()
            return True
        else:
            return False


class ClientCredentialsTokenProvider(TokenProvider):
    def __init__(self, token_file, config):
        self._config = config
        self._token_file = token_file
        try:
            self._oauth2_config = load_oauth2_access_token(self._token_file)
            if 'access_token' in self._oauth2_config:
                self._access_token = self._oauth2_config['access_token']
            else:
                raise RawException('No access_token field found in {}'.format(self._token_file))
        except RawException as ex:
            self._oauth2_config = None
            self._access_token = None

    def __do_client_credentials(self):
        logger.debug("Exchanging client credentials for token.")
        domain = self._config.oauth2_domain
        client_id = self._config.oauth2_client_id
        client_secret = self._config.oauth2_client_secret
        audience = self._config.oauth2_audience
        provider = self._config.oauth2_provider
        missing_param = []
        if domain is None:
            missing_param.append('domain')
        if client_id is None:
            missing_param.append('client_id')
        if client_secret is None:
            missing_param.append('client_secret')
        if provider is None:
            missing_param.append('provider')

        if len(missing_param) != 0:
            raise RawException('Missing parameters from configuration: {}.\n'
                            'For the client credentials flow, please make sure the following properties are correctly set in your configuration file:\n'
                            '[auth_oauth2]\n'
                            'provider = <provider>\n'
                            'grant_type = client_credentials\n'
                            'domain = <domain>\n'
                            'client_id = <id>\n'
                            'client_secret = <secret>\n'.format(', '.join(missing_param)))

        data = {
            'client_id': client_id,
            'client_secret': client_secret,
            'grant_type': 'client_credentials'
        }

        if provider == 'auth0':
            if audience is None:
                raise RawException('Configuration property "auth_oauth2.audience" must be specified when using client credentials with Auth0 provider.')
            data['audience'] = audience

        headers = {'Accept': 'application/json'}
        url = 'https://{}/oauth/token'.format(domain)
        logger.debug('Obtaining OAuth2 token for audience {}. URL: {}'.format(audience, url))
        resp = requests.post(url, data=data, headers=headers, proxies=get_requests_proxy(self._config))
        if resp.status_code in [400, 401, 403, 404]:
            error = json.loads(resp.text)
            raise RawException('Could not exchange client credentials for access token.\nError: {}'.format(error))
        elif resp.status_code != 200:
            raise RawException(
                'Unexpected response.\nResponse status: {}\nResponse body: {}'.format(resp.status_code, resp.text))
        else:
            save_oauth2_access_token(self._token_file, resp.json())
            self._oauth2_config = load_oauth2_access_token(self._token_file)
            if 'access_token' in self._oauth2_config:
                self._access_token = self._oauth2_config['access_token']
            else:
                raise RawException('No access_token field found in {}'.format(self._token_file))

    def get_token(self):
        if not self._access_token:
            self.__do_client_credentials()
        return self._access_token

    def refresh_token(self):
        self.__do_client_credentials()
        return True


class RawException(Exception):
    """Top-level RAW exception."""

    def __init__(self, msg):
        Exception.__init__(self, msg)


class APIError(RawException):
    """Exception thrown when an API call failed."""

    def __init__(self, error):
        RawException.__init__(self, 'API error: {}'.format(error['message']))
        self.code = error['code']
        self.message = error['message']


class Unauthorized(RawException):
    """Exception thrown when the client credentials are refused by the service"""

    def __init__(self, error):
        RawException.__init__(self, 'Unauthorized: {}'.format(error))


class ConfigurationException(RawException):
    """Exception thrown when the configuration is missing or incorrect."""

    def __init__(self, msg):
        RawException.__init__(self, msg)


class QueryValidationError(RawException):
    """Exception thrown when a query failed during validation."""

    def __init__(self, errors):
        RawException.__init__(self, 'Invalid query')
        self.validation_errors = errors

    def __str__(self):
        validation_errors_list = []
        for error in self.validation_errors:
            msg = error['message']
            positions = []
            for p in error['positions']:
                begin = p['begin']
                end = p['end']
                positions.append('{}:{} to {}:{}'.format(begin['line'], begin['column'], end['line'], end['column']))
            if positions:
                validation_errors_list.append('- {}. Positions: {}'.format(msg, '; '.join(positions)))
            else:
                validation_errors_list.append('- {}.'.format(msg))
        return 'Invalid query:\n{}'.format('\n'.join(validation_errors_list))


class QueryExecutionError(RawException):
    """Exception thrown when a query fails to execute."""

    def __init__(self, error):
        RawException.__init__(self, 'Execution error')
        self.error = error

    def __str__(self):
        return 'Execution error: {}'.format(self.error)


def get_requests_proxy(config):
    if config.http_proxy:
        proxies = {"http": config.http_proxy, "https": config.https_proxy}
    else:
        proxies = None
    return proxies


class Raw:
    """
    Raw client class
    """
    def __init__(self, executor_url, creds_url, kafka_url, storage_url,  token_provider, config, **kwargs):
        self.executor_url = executor_url
        self.creds_url = creds_url
        self.kafka_url = kafka_url
        self.storage_url = storage_url
        self.token_provider = token_provider
        self.verbose = kwargs.get('verbose', False)
        self.save = kwargs.get('save', False)
        self.output_dir = kwargs.get('output_dir')
        if not config.verify_tls_cert:
            urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
        self.config = config

        self.request_headers = {}
        if self.output_dir is None:
            self.output_dir = tempfile.gettempdir()

    def _gen_query_name(self):
        if not os.path.exists(self.output_dir):
            os.makedirs(self.output_dir)
        existing_files = os.listdir(self.output_dir)
        for i in itertools.count():
            name = 'query{:0>5d}'.format(i)
            exists = any(filename.startswith(name) for filename in existing_files)
            if not exists:
                return os.path.join(self.output_dir, name)

    def _save_as_curl(self, request, path):
        curl_lines = ['curl -X {} {}'.format(request.method.upper(), request.url)]
        # Send only these headers:
        copied_headers = ['Authorization', 'Content-Type', 'Accept-Encoding']
        for k, v in request.headers.items():
            if k in copied_headers:
                curl_lines.append('  -H "{}: {}"'.format(k, v))
        if request.body is not None:
            curl_lines.append('  -d "{}"'.format(request.body))
        with open(path, 'w') as f:
            f.write('#!/bin/bash\n')
            f.write('\\\n'.join(curl_lines))
        os.chmod(path, 0o755)

    def _save_request(self, request, path):
        with open(path, 'w') as f:
            f.write(request.method + ' ' + request.url)
            f.write('\n'.join(['{}: {}'.format(k, v) for k, v in request.headers.items()]))
            f.write('\n\n')

            if request.body is not None:
                f.write(str(request.body))

    def _save_response(self, response, path):
        with open(path, 'wb') as f:
            f.write(response.content)

    def _save(self, response):
        if self.save:
            query_file = self._gen_query_name()
            logger.debug('Writing request, response and curl script as %s', query_file)
            self._save_request(response.request, query_file + '.in')
            self._save_response(response, query_file + '.out')
            self._save_as_curl(response.request, query_file + '.curl.sh')

    def _call_service(self, method, url, headers, attempt=1, **kwargs):
        request_headers = _COMMON_HEADERS.copy()
        request_headers['Authorization'] = 'Bearer {}'.format(self.token_provider.get_token())
        if headers is not None:
            request_headers.update(headers)
        try:
            session = requests.session()
            session.keep_alive = False  # TODO True?
            response = session.request(method, url, headers=request_headers,
                                       timeout=(self.config.connect_timeout, self.config.read_timeout),
                                       verify=self.config.verify_tls_cert, proxies=get_requests_proxy(self.config),
                                       **kwargs)
        except requests.exceptions.ConnectionError as e:
            # The requests error message is extremely confusing in this common case...
            raise RawException('Service unavailable: connection to {} failed.\nError: {}'.format(url, e))
        if response.status_code in [401, 403]:
            # If using client credentials, try to obtain a new access token.
            # if self.config.oauth2_grant_type == 'authorization_code':
            #     logger.warn('Authorization error contacting server, access token may be invalid or expired. Please rerun "rawcli login" to obtain a new token')
            if (isinstance(self.token_provider, ClientCredentialsTokenProvider) or isinstance(self.token_provider, AuthorizationCodeTokenProvider) or isinstance(self.token_provider, JsonTokenProvider)) and attempt == 1:
                logger.debug('Authorization error contacting server, access token may be invalid or expired. Trying to obtain a new token')
                if self.token_provider.refresh_token():
                    return self._call_service(method, url, headers, attempt+1, **kwargs)
                else:
                    raise RawException('No refresh token available. Please obtain a new access token (rawcli login)')
        self._save(response)
        return response

    def _call_executor(self, method, path, headers=None, **kwargs):
        if not self.executor_url:
            raise RawException('Executor service endpoint is not configured')
        return self._call_service(method, self.executor_url + path, headers, **kwargs)

    def _call_creds(self, method, path, headers=None, **kwargs):
        if not self.creds_url:
            raise RawException('Credentials service endpoint is not configured')
        return self._call_service(method, self.creds_url + path, headers, **kwargs)

    def _call_kafka(self, method, path, headers=None, **kwargs):
        if not self.kafka_url:
            raise RawException('Kafka service endpoint is not configured')
        return self._call_service(method, self.kafka_url + path, headers, **kwargs)

    def _call_storage(self, method, path, headers=None, **kwargs):
        if not self.storage_url:
            raise RawException('Storage service endpoint is not configured')
        return self._call_service(method, self.storage_url + path, headers, **kwargs)

    @staticmethod
    def _handle_error_response(resp):
        if resp.status_code == 400 and resp.text:
            json_resp = resp.json()
            if 'validationErrors' in json_resp:
                raise QueryValidationError(json_resp['validationErrors'])
            elif 'runtimeError' in json_resp:
                raise QueryExecutionError(json_resp['runtimeError'])
            else:
                raise APIError(json_resp)
        elif resp.status_code in [401, 403] and resp.text:
            raise Unauthorized(resp.json()['message'])
        elif resp.status_code == 503:
            raise RawException(
                'Service unavailable. Please try again later.\nResponse status: {}\nResponse body: {}'.format(
                    resp.status_code, resp.text))
        else:
            raise RawException(
                'Unexpected response.\nResponse status: {}\nResponse body: {}'.format(resp.status_code, resp.text))

    # ====================
    # Queries
    # ====================
    def query_validate(self, query):
        """Validates a query.

        Args:
            query (str): Query to validate.
        Returns:
            dict: example::

                {
                  'type': 'collection(int)',
                  'errors': [],
                  'comment': None,
                  'declarations': {}
                }
        """
        resp = self._call_executor('POST', '/1/public/query-validate', data=query)
        if resp.status_code == 200:
            response = resp.json()
            response['errors'] = []
            return response
        else:
            self._handle_error_response(resp)

    def query(self, query, msg_callback=None, np_arrays=False, use_relativedeltas=True, with_type=False, as_sql=False, log_level=None):
        """Executes a query returning python objects.

        Args:
            query (str): Query string to run on the server
            msg_callback (function(str, str)): Callback function for messages in the query stream args (level, msg)
            np_arrays (bool): If true convert arrays to numpy.arrays
            use_relativedeltas (bool): If true intervals are converted to dateutil.relativedeltas instead of datetime.timedelta
            with_type (bool): If true returns a tuple (data, tipe)
            as_sql (bool): If true uses SQL compatibility mode.
            log_level (str): log level of the query logger.

        Returns:
            A generator of python objects if it is a collection or the object itself if something else.
            or a tuple (generator, dict) if with_type == true
        """

        # The encoding is always utf-8, a config in the server, a item in the header?
        # TODO: Find out where the encoding is set
        stream = self.query_as_text(query, output_format='raw', encoding='utf-8', msg_callback=msg_callback, as_sql=as_sql, log_level=log_level)
        try:
            tipe = json.loads(stream.readline(), object_pairs_hook=OrderedDict)
        except (KeyboardInterrupt, SystemExit, Exception):
            # Close stream if query cancelled while waiting for first line w/ type information
            stream.close()
            raise

        if tipe['type'] == 'collection' and np_arrays:
            import numpy as np
            data = np.array([x for x in RawPythonIterator(stream, tipe, np_arrays, use_relativedeltas)])
        elif tipe['type'] == 'collection':
            data = RawPythonIterator(stream, tipe, np_arrays, use_relativedeltas)
        else:
            try:
                obj = json.loads(stream.readline(), object_pairs_hook=OrderedDict)
                data = convert_types(obj, tipe, np_arrays, use_relativedeltas)
            finally:
                stream.close()
        if with_type:
            return (data, tipe)
        else:
            return data

    def query_as_text(self, query, output_format, encoding='utf-8', msg_callback=None, as_sql=False, log_level=None):
        """Executes a query returning a text stream.

        Args:
            query (str): The query string to execute on the server.
            ouput_format (str): Text output format: 'json', 'hjson' or 'csv'.
            encoding (str): Output encoding,  defaults to 'utf-8'.
            msg_callback function(str, str): Callback function for messages in the query stream args (level, msg).
            as_sql (bool): If true uses SQL compatibility mode.
            log_level (str): log level of the query logger.

        Returns:
            TextIOBase: Stream with the data
        """
        if output_format not in QUERY_TEXT_FORMATS:
            raise RawException(
                'Invalid output format for text output. Available formats are: {}'.format(QUERY_TEXT_FORMATS))
        stream = self.query_as(query, output_format, msg_callback=msg_callback, as_sql=as_sql, log_level=log_level)
        return RawTextIOWrapper(stream, encoding)

    def query_as(self, query, output_format, msg_callback=None, windows_line_ending=None, as_sql=False, log_level=None):
        """Executes a query returning a binary stream.
        The byte stream object includes a field 'query_id' with the unique query identifier.

        Args:
            query (str): The query string to execute on the server.
            ouput_format (str):  Output data format: 'json', 'hjson', 'csv', 'sqlite', 'parquet-avro' or 'parquet-raw'.
            msg_callback function(str, str): Callback function for messages in the query stream args (level, msg).
            windows_line_ending (bool): If true uses Windows-style line-ending '\\r\\n'.
            as_sql (bool): If true uses SQL compatibility mode.
            log_level (str): log level of the query logger.

        Returns:
            IOBase: Stream with the data
        """
        if output_format not in QUERY_FORMATS:
            raise RawException('Invalid output format. Available formats are: {}'.format(QUERY_FORMATS))

        if not log_level:
            log_level = self.config.query_log_level.lower()
        else:
            log_level = log_level.lower()

        if log_level not in QUERY_LOG_LEVELS:
            raise RawException('Unsupported query_log_level "%s" supported values are: %s' % (log_level, ", ".join(QUERY_LOG_LEVELS)))

        if self.config.query_response_transfer == 'http':
            if log_level != 'none' and msg_callback is not None:
                raise RawException(
                    'The "query_response_transfer = http" option is not compatible with query log messages.\n'
                    'Please update your raw.ini or pass "msg_callback=None" to query methods.')
            if as_sql:
                raise RawException('SQL compatibility mode only supported with "query_response_transfer = websockets"')

            return self.__query_stream(query, output_format, windows_line_ending)
        elif self.config.query_response_transfer == 'websockets':
            return self.__query_web_sockets(query, output_format, log_level, msg_callback, windows_line_ending, as_sql)
        else:
            raise RawException('Unsupported query_response_transfer type: {}'.format(self.config.query_response_transfer))

    def __query_web_sockets(self, query, output_format, log_level, msg_callback, windows_line_ending, as_sql):
        options = {_RAW_OUTPUT_FORMAT: output_format}
        if windows_line_ending:
            options[_RAW_WINDOWS_LINE_ENDING] = "true"
        return self.__try_query_web_sockets(query, options, log_level, msg_callback, as_sql)

    def __try_query_web_sockets(self, query, options, log_level, msg_callback, as_sql, attempt=1):
        logger.debug("Trying to connect to {}. Attempt #{}".format(self.executor_url, attempt))
        token = self.token_provider.get_token()
        try:
            return WebSocketInputStream(self.executor_url, token, query, options, log_level, msg_callback, as_sql, self.config)
        except Unauthorized as ex:
            if (isinstance(self.token_provider, ClientCredentialsTokenProvider) or isinstance(self.token_provider, AuthorizationCodeTokenProvider) or isinstance(self.token_provider, JsonTokenProvider)) and attempt == 1:
                logger.debug('Authorization error contacting server, access token may be invalid or expired. Trying to obtain a new token')
                if self.token_provider.refresh_token():
                    return self.__try_query_web_sockets(query, options, log_level, msg_callback, as_sql, attempt+1)
                else:
                    raise RawException('No refresh token available. Please obtain a new access token (rawcli login)')
            raise ex

    def __query_stream(self, query, output_format, windows_line_ending):
        """Execute a query. The response can either be a plain HTTP response (with Content-Length set)
          or a chunked response, if the size of the response exceeds a certain value (typically 16KB).

        Returns:
            a generator whose elements are the chunks received from the server.
            This method does not transform or interpret the response in any way.
        """
        options = {_RAW_OUTPUT_FORMAT: output_format}
        if windows_line_ending:
            options[_RAW_WINDOWS_LINE_ENDING] = "true"
        headers = {"X-RAW-{}".format(k): options[k] for k in options}
        resp = self._call_executor('POST', '/1/public/query', headers=headers, data=query, stream=True)
        if resp.status_code == 200:
            return HttpInputStream(resp)
        else:
            self._handle_error_response(resp)

    # ====================
    # S3 Buckets
    # ====================
    def buckets_list(self):
        """Lists S3 buckets registered in the credentials server.
        """
        resp = self._call_creds('GET', '/1/public/s3buckets')
        if resp.status_code == 200:
            return resp.json()
        else:
            self._handle_error_response(resp)

    def buckets_show(self, name):
        """Shows details for an S3 bucket registered in the credentials server. """
        resp = self._call_creds('GET', '/1/public/s3buckets/{}'.format(name))
        if resp.status_code == 200:
            return resp.json()
        else:
            self._handle_error_response(resp)

    def buckets_register(self, name, region=None, access_key=None, secret_key=None):
        """Registers an S3 bucket in the credentials server.

        Args:
            name (str): Name of the bucket
            region (str): Region of the bucket
            access_key (str): Access key of the bucket
            secret_key (str): Secret key of the bucket
        """
        json_body = {'name': name}
        if region is not None:
            json_body['region'] = region
        if access_key is not None:
            json_body['credentials'] = {'accessKey': access_key, 'secretKey': secret_key}
        resp = self._call_creds('POST', '/1/public/s3buckets', json=json_body)
        if resp.status_code != 201:
            self._handle_error_response(resp)

    def buckets_unregister(self, name):
        """Unregisters an S3 bucket from the credentials server."""
        resp = self._call_creds('DELETE', '/1/public/s3buckets/{}'.format(name))
        if resp.status_code != 204:
            self._handle_error_response(resp)

    # ====================
    # HDFS delegation tokens
    # ====================
    def hdfs_token_show(self):
        """Shows registered HDFS token"""
        resp = self._call_creds('GET', '/1/public/hdfs-tokens')
        if resp.status_code == 200:
            return resp.json()
        else:
            self._handle_error_response(resp)

    def hdfs_token_register(self, token):
        """Registers an hdfs delegation token in the credentials server.

        Args:
            token (str): The token to register.
        """
        resp = self._call_creds('POST', '/1/public/hdfs-tokens', json={'token': token})
        if resp.status_code != 201:
            self._handle_error_response(resp)

    def hdfs_token_unregister(self):
        """Unregisters an hdfs delegation token from the credentials server.
        """
        resp = self._call_creds('DELETE','/1/public/hdfs-tokens')
        if resp.status_code != 204:
            self._handle_error_response(resp)

    # ====================
    # Http Auth
    # ====================
    def http_auth_list(self):
        """Lists http credentials/authorization methods registered in the credentials server.
        """
        resp = self._call_creds('GET', '/1/public/http-creds')
        if resp.status_code == 200:
            return resp.json()
        else:
            self._handle_error_response(resp)

    def http_auth_show(self, url):
        """Shows details for an http credentials registered in the credentials server.

        Args:
            url (string): url to show the info.

        Returns:
            dict: example::

            {
                'credentials': {'type': 'basic-auth',
                                'username': 'john',
                                'password': 'secret'},
                'prefixUrl': 'http://somewhere.com'
            }

        """
        resp = self._call_creds('GET', '/1/public/http-creds', params=dict(name=url))
        if resp.status_code == 200:
            return resp.json()
        else:
            self._handle_error_response(resp)

    def http_auth_register(self, url, credentials):
        """Registers an http credentials in the credentials server.

        Args:
            credentials (dict): Credentials to register,
                example by authorization type::

                    bearer token:  { 'token': 'my_token', 'refreshToken': None,
                                    'tokenUrl': None }
                    client secret: { 'clientId': 'my_id',
                                        'clientSecret': 'my_secret',
                                        'tokenUrl': 'http://token-service.com',
                                        'useBasicAuth': None   }
                    basic auth:    {'user': 'john', 'password': 'passwd'}

        """
        if 'token' in credentials:
            credentials['type'] = 'oauth-token'
        elif 'clientId' in credentials:
            credentials['type'] = 'oauth-client'
            if not 'useBasicAuth' in credentials:
                credentials['useBasicAuth'] = True
        elif 'password' in credentials:
            credentials['type'] = 'basic-auth'
        else:
            raise RawException('Unknown credentials type: {}'.format(credentials))

        json_body = {'prefixUrl': url, 'credentials': credentials}

        resp = self._call_creds('POST', '/1/public/http-creds', json=json_body)
        if resp.status_code != 201:
            self._handle_error_response(resp)

    def http_auth_unregister(self, url):
        """Unregisters http credentials from the credentials server.

        Args:
            url (str): Url to register, e.g.: 'http://somewhere.com'
        """
        resp = self._call_creds('DELETE', '/1/public/http-creds', params=dict(name=url))
        if resp.status_code != 204:
            self._handle_error_response(resp)

    # ====================
    # RDBMS
    # ====================
    def rdbms_list(self):
        """Lists relational databases registered in the credentials server.
        """
        resp = self._call_creds('GET', '/1/public/rdbmsservers')
        if resp.status_code == 200:
            return resp.json()
        else:
            self._handle_error_response(resp)

    def rdbms_show(self, name):
        """Shows details for a relational database registered in the credentials server.
        """
        resp = self._call_creds('GET', '/1/public/rdbmsservers/{}'.format(name))
        if resp.status_code == 200:
            return resp.json()
        else:
            self._handle_error_response(resp)

    def rdbms_register_postgresql(self, name, host, database, port=None, username=None, password=None):
        """Registers a PostgreSQL database in the credentials server.

        Args:
            name (str): Name of the server to be used in Raw.
            host (str): Hostname of the database server.
            database (str): Database to connect to.
            port (int): Port number of the server. Defaults to None.
            username (str): Username to login in the server. Defaults to None.
            password (str): User password to login in the server. Defaults to None.

        Example::

            client = new_raw_client()
            client.rdbms_register_postgresql('my_server', 'db_server.somewhere.com', 'my_db')
        """
        database = {'type': 'postgresql', 'host': host, 'database': database}
        if username is not None:
            database['username'] = username
            database['password'] = password
        if port is not None:
            if not isinstance(port, int):
                raise RawException('port is invalid')
            database['port'] = port
        json_body = {'name': name, 'database': database}
        resp = self._call_creds('POST', '/1/public/rdbmsservers', json=json_body)
        if resp.status_code != 201:
            self._handle_error_response(resp)

    def rdbms_register_mysql(self, name, host, database, port=None, username=None, password=None):
        """Registers a MySQL database in the credentials server.

        Args:
            name (str): Name of the server to be used in Raw.
            host (str): Hostname of the database server.
            database (str): Database to connect to.
            port (int): Port number of the server. Defaults to None.
            username (str): Username to login in the server. Defaults to None.
            password (str): User password to login in the server. Defaults to None.

        Example::

            client = new_raw_client()
            client.rdbms_register_mysql('my_server', 'db_server.somewhere.com', 'clients_db')
        """
        database = {'type': 'mysql', 'host': host, 'port': port, 'database': database}
        if username is not None:
            database['username'] = username
            database['password'] = password
        if port is not None:
            database['port'] = port
        json_body = {'name': name, 'database': database}
        resp = self._call_creds('POST', '/1/public/rdbmsservers', json=json_body)
        if resp.status_code != 201:
            self._handle_error_response(resp)

    def rdbms_register_sqlserver(self, name, host, database, port=None, username=None, password=None):
        """Registers a Microsoft SQL Server database in the credentials server.

        Args:
            name (str): Name of the server to be used in Raw.
            host (str): Hostname of the database server.
            database (str): Database to connect to.
            port (int): Port number of the server. Defaults to None.
            username (str): Username to login in the server. Defaults to None.
            password (str): User password to login in the server. Defaults to None.

        Example::

            client = new_raw_client()
            client.rdbms_register_sqlserver('my_server', 'db_server.somewhere.com', 'clients_db')
        """
        database = {'type': 'sqlserver', 'host': host, 'port': port, 'database': database}
        if username is not None:
            database['username'] = username
            database['password'] = password
        if port is not None:
            database['port'] = port
        json_body = {'name': name, 'database': database}
        resp = self._call_creds('POST', '/1/public/rdbmsservers', json=json_body)
        if resp.status_code != 201:
            self._handle_error_response(resp)

    def rdbms_register_oracle(self, name, host, database, port=None, username=None, password=None):
        """Registers a Oracle database in the credentials server.

        Args:
            name (str): Name of the server to be used in Raw.
            host (str): Hostname of the database server.
            database (str): Database to connect to.
            port (int): Port number of the server. Defaults to None.
            username (str): Username to login in the server. Defaults to None.
            password (str): User password to login in the server. Defaults to None.

        Example::

            client = new_raw_client()
            client.rdbms_register_oracle('my_server', 'db_server.somewhere.com', 'clients_db')
        """
        database = {'type': 'oracle', 'host': host, 'port': port, 'database': database}
        if username is not None:
            database['username'] = username
            database['password'] = password
        if port is not None:
            database['port'] = port
        json_body = {'name': name, 'database': database}
        resp = self._call_creds('POST', '/1/public/rdbmsservers', json=json_body)
        if resp.status_code != 201:
            self._handle_error_response(resp)


    def rdbms_register_teradata(self, name, host, port=None, username=None, password=None, parameters={}):
        """Registers a Teradata database in the credentials server.

        Args:
            name (str): Name of the server to be used in Raw.
            host (str): Hostname of the database server.
            database (str): Database to connect to.
            port (int): Port number of the server. Defaults to None.
            username (str): Username to login in the server. Defaults to None.
            password (str): User password to login in the server. Defaults to None.
            parameters (dict): Extra parameters to be passed to the teradata server on the connection
                    e.g.: {'CHARSET': 'UTF8'}

        Example::

            client = new_raw_client()
            client.rdbms_register_teradata('my_server', 'db_server.somewhere.com', parameters={'TMODE': 'ANSI'})
        """
        database = {'type': 'teradata', 'host': host, 'port': port, 'parameters': parameters}
        if username is not None:
            database['username'] = username
            database['password'] = password
        if port is not None:
            database['port'] = port
        json_body = {'name': name, 'database': database}
        resp = self._call_creds('POST', '/1/public/rdbmsservers', json=json_body)
        if resp.status_code != 201:
            self._handle_error_response(resp)


    def rdbms_unregister(self, name):
        """Unregisters a relational database from the credentials server.

        Example::
            client = new_raw_client()
            client.rdbms_unregister('oracle1')
        """
        resp = self._call_creds('DELETE', '/1/public/rdbmsservers/{}'.format(name))
        if resp.status_code != 204:
            self._handle_error_response(resp)

    # ====================
    # Tables
    # ====================
    def tables_list_names(self):
        """Lists table names.
        """
        resp = self._call_executor('GET', '/1/public/tables')
        if resp.status_code == 200:
            names = [u['name'] for u in resp.json()]
            return names
        else:
            self._handle_error_response(resp)

    def tables_list(self):
        """Lists tables.

        Returns:
            list: List with dicts of table-info see `tables_show` for the format.
        """
        resp = self._call_executor('GET', '/1/public/tables')
        if resp.status_code == 200:
            return resp.json()
        else:
            self._handle_error_response(resp)

    def tables_show(self, name):
        """Shows details of a table.

        Returns:
            dict: example::

                {
                  'name': 'hello',
                  'type': 'collection(record(a: int, b: string))',
                  'location': 'hdfs://nuc1:9000/tmp/hello',
                  'createdAt': [2020, 9, 7, 19, 4, 51, 732000000],
                  'updatedAt': None
                }
        """
        resp = self._call_executor('GET', '/1/public/tables/{}'.format(name))
        if resp.status_code == 200:
            return resp.json()
        else:
            self._handle_error_response(resp)

    def tables_create(self, name, tipe, partition=[], indexes=[], location=None):
        """Creates a table.

        Args:
            name (str): Name of the table to create.
            tipe (str): Type of the query, e.g.: 'collection(record(a: int, b: string))'
            partition (list): Fields to partition table. Defaults to [],  e.g.: ['a']
            indexes (list): Indexes to add to the table. Only applicable for external table. Defaults to [],  e.g.: ['a,b', 'a|b']
            location (str): None (default) to create an internal table. Otherwise, pass the physical location of the external table,  e.g.: 'hdfs:///some/path'.
        """
        if location is None:
            if indexes:
                raise RawException(
                    'Cannot create internal table with indexes. Add indexes separately.')
            resp = self._call_executor('PUT', '/1/public/tables',
                                       json={'table': 'internal', 'name': name, 'type': tipe, 'partition': partition})
        else:
            resp = self._call_executor('PUT', '/1/public/tables',
                                       json={'table': 'external', 'name': name, 'type': tipe, 'partition': partition, 'indexes': indexes, 'location': location})
        if resp.status_code == 200:
            return
        else:
            self._handle_error_response(resp)

    def tables_drop(self, name):
        """Drops a table."""
        resp = self._call_executor('DELETE', '/1/public/tables/{}'.format(name))
        if resp.status_code == 204:
            return
        else:
            self._handle_error_response(resp)

    def tables_optimize(self, name):
        """Optimizes a table.
        """
        resp = self._call_executor('POST', '/1/public/tables/{}'.format(name), json={'action': 'optimize'})
        if resp.status_code == 200:
            return
        else:
            self._handle_error_response(resp)

    def tables_vacuum(self, name, hours):
        """Vacuums a table.
        """
        resp = self._call_executor('POST', '/1/public/tables/{}'.format(name), json={'action': 'vacuum', 'hours': hours})
        if resp.status_code == 200:
            return
        else:
            self._handle_error_response(resp)

    def tables_insert(self, name, data):
        """Insert data into a table.
        """
        resp = self._call_executor('PUT', '/1/public/tables/{}'.format(name), data=data)
        if resp.status_code == 200:
            return
        else:
            self._handle_error_response(resp)

    # ====================
    # Table Indexes
    # ====================
    def table_indexes_list(self, name):
        """Lists table indexes.

        Args:
            name: Name of the table.

        Returns:
            list: List of indexes, example::

                [
                    {'index': {'columns': ['aaa', 'bbb', 'ccc'], 'type': 'and'},
                    'createdAt': [2020, 9, 23, 15, 4, 21, 928000000],
                    'updatedAt': None}
                ]

        """
        resp = self._call_executor('GET', '/1/public/table_indexes/{}'.format(name))
        if resp.status_code == 200:
            return resp.json()
        else:
            self._handle_error_response(resp)

    def table_indexes_create(self, name, index):
        """Creates a table index.

        Args:
            name (str): Table name.
            index (str): Index to create:

                =========== ===================================
                And Index   Comma separated list of fields
                                example: 'a,b'
                Or Index    Pipe separated, '|' list of fields
                                example:  'a|b'
                =========== ===================================
        """
        resp = self._call_executor('PUT', '/1/public/table_indexes/{}'.format(name), json=index)
        if resp.status_code == 200:
            return
        else:
            self._handle_error_response(resp)

    def table_indexes_drop(self, name, index):
        """Drops a table index.
        """
        resp = self._call_executor('DELETE', '/1/public/table_indexes/{}'.format(name), json=index)
        if resp.status_code == 204:
            return
        else:
            self._handle_error_response(resp)

    def table_indexes_optimize(self, name, index):
        """Optimizes a table index.
        """
        resp = self._call_executor('POST', '/1/public/table_indexes/{}'.format(name), json={'action': 'optimize', 'index': index})
        if resp.status_code == 200:
            return
        else:
            self._handle_error_response(resp)

    def table_indexes_vacuum(self, name, index, hours):
        """Vacuums a table index.
        """
        resp = self._call_executor('POST', '/1/public/table_indexes/{}'.format(name), json={'action': 'vacuum', 'index': index})
        if resp.status_code == 200:
            return
        else:
            self._handle_error_response(resp)

    # ====================
    # Views
    # ====================
    def views_validate(self, view):
        """Validate a view.

        Returns:
            dict: example::

                {
                    'comment': None,
                    'type': 'collection(record(a: int, b: string))',
                    'declarations': {},
                    'errors': []
                }

        """
        resp = self._call_executor('GET', '/1/public/view-validate/{}'.format(view))
        if resp.status_code == 200:
            response = resp.json()
            response['errors'] = []
            return response
        else:
            self._handle_error_response(resp)

    def views_list_names(self):
        """Lists view names.

        Returns:
            list: example::
                ['view1', 'view2']
        """
        resp = self._call_executor('GET', '/1/public/views')
        if resp.status_code == 200:
            names = [u['name'] for u in resp.json()]
            return names
        else:
            self._handle_error_response(resp)

    def views_list(self):
        """Lists views.

        Returns:
            list: List of dicts with views.
        """
        resp = self._call_executor('GET', '/1/public/views')
        if resp.status_code == 200:
            return resp.json()
        else:
            self._handle_error_response(resp)

    def views_show(self, name):
        """Shows details for a view.

        Returns:
            dict: example::

            {
                'name': 'my_view',
                'query': 'select * from read("file://path/file.json")',
                'createdAt': [2020, 9, 22, 11, 4, 28, 309000000],
                'updatedAt': None
             }
        """
        resp = self._call_executor('GET', '/1/public/views/{}'.format(name))
        if resp.status_code == 200:
            return resp.json()
        else:
            self._handle_error_response(resp)

    def views_create(self, name, query):
        """Creates a view.

        Args:
            name (str): Name of the package.
            query (str): Raw query of the view.
        """
        resp = self._call_executor('PUT', '/1/public/views/{}'.format(name), data=query)
        if resp.status_code == 200:
            return
        else:
            self._handle_error_response(resp)

    def views_drop(self, name):
        """Drops a view."""
        resp = self._call_executor('DELETE', '/1/public/views/{}'.format(name))
        if resp.status_code == 204:
            return
        else:
            self._handle_error_response(resp)

    # ====================
    # Materialized Views
    # ====================
    def materialized_views_validate(self, view):
        """Validate a materialized view.

        Returns:
            dict: example::

                {
                    'comment': None,
                    'type': 'collection(record(a: int, b: string))',
                    'declarations': {},
                    'errors': [],
                } """
        resp = self._call_executor('GET', '/1/public/materialized-view-validate/{}'.format(view))
        if resp.status_code == 200:
            response = resp.json()
            response['errors'] = []
            return response
        else:
            self._handle_error_response(resp)

    def materialized_views_list_names(self):
        """Lists materialized view names.

        Returns:
            list: example::
                ['view1', 'view2']
        """
        resp = self._call_executor('GET', '/1/public/materialized-views')
        if resp.status_code == 200:
            names = [u['name'] for u in resp.json()]
            return names
        else:
            self._handle_error_response(resp)

    def materialized_views_list(self):
        """Lists materialized views.
        """
        resp = self._call_executor('GET', '/1/public/materialized-views')
        if resp.status_code == 200:
            return resp.json()
        else:
            self._handle_error_response(resp)

    def materialized_views_show(self, name):
        """Shows details for a materialized view.

        Returns:
            dict: example::

            {
                'name': 'my_view',
                'query': 'select * from read("file://path/file.json")',
                'createdAt': [2020, 9, 22, 11, 4, 28, 309000000],
                'updatedAt': None
            }
        """
        resp = self._call_executor('GET', '/1/public/materialized-views/{}'.format(name))
        if resp.status_code == 200:
            return resp.json()
        else:
            self._handle_error_response(resp)

    def materialized_views_create(self, name, query):
        """Creates a materialized view.

        Args:
            name (str): Name of the package.
            query (str): Raw query of the view to create.
        """
        resp = self._call_executor('PUT', '/1/public/materialized-views/{}'.format(name), data=query)
        if resp.status_code == 200:
            return
        else:
            self._handle_error_response(resp)

    def materialized_views_drop(self, name):
        """Drops a materialized view."""
        resp = self._call_executor('DELETE', '/1/public/materialized-views/{}'.format(name))
        if resp.status_code == 204:
            return
        else:
            self._handle_error_response(resp)

    # ====================
    # Packages
    # ====================
    def packages_validate(self, pkg):
        """Validate a package.

        Returns:
            dict: example::

                {
                  'type': 'collection(int)',
                  'errors': [],
                  'comment': None,
                  'declarations': {}
                }
        """
        resp = self._call_executor('GET', '/1/public/package-validate/{}'.format(pkg))
        if resp.status_code == 200:
            response = resp.json()
            response['errors'] = []
            return response
        else:
            self._handle_error_response(resp)

    def packages_list_names(self):
        """Lists package names.

        Returns:
            list: example::
                ['pkg1', 'pkg2']
        """
        resp = self._call_executor('GET', '/1/public/packages')
        if resp.status_code == 200:
            names = [u['name'] for u in resp.json()]
            return names
        else:
            self._handle_error_response(resp)

    def packages_list(self):
        """Lists packages.
        """
        resp = self._call_executor('GET', '/1/public/packages')
        if resp.status_code == 200:
            return resp.json()
        else:
            self._handle_error_response(resp)

    def packages_show(self, name):
        """Shows details for a package.

        Returns:
            dict: example::

                {
                  'name': 'test_pkg',
                  'query': '// my cool package
                            foo(a: int) := a + 1;
                            foo(1);',
                  'createdAt': [2020, 10, 12, 15, 29, 6, 723000000],
                  'updatedAt': None
                }
        """
        resp = self._call_executor('GET', '/1/public/packages/{}'.format(name))
        if resp.status_code == 200:
            return resp.json()
        else:
            self._handle_error_response(resp)

    def packages_create(self, name, query):
        """Creates a package.

        Args:
            name (str): Name of the package to create.
            query (str): Raw code of the package to create.
        """
        resp = self._call_executor('PUT', '/1/public/packages/{}'.format(name), data=query)
        if resp.status_code == 200:
            return
        else:
            self._handle_error_response(resp)

    def packages_drop(self, name):
        """Drops a package. """
        resp = self._call_executor('DELETE', '/1/public/packages/{}'.format(name))
        if resp.status_code == 204:
            return
        else:
            self._handle_error_response(resp)

    # ====================
    # Kafka
    # ====================
    def kafka_create(self, name, host, port, topic, interval, location, transform, indexes=[]):
        """Creates a Kafka table.

        Args:
            name (str): Name of kafka table to create.
            host (str): Hostname of the kafka server.
            port (int): Port number of the kafka server.
            topic (str): Kafka topic to listen on.
            interval (datetime.timedelta): period of time to poll the kafka server for new data (max resolution is 1 ms).
            location (str): Place to store the data e.g.: 'hdfs://node/path/sub-path'
        """
        interval_millis = int(interval.total_seconds() * 1000)
        request = dict(host=host, port=port, topic=topic, transform=transform, intervalMillis=interval_millis, indexes=indexes, location=location)
        resp = self._call_kafka('PUT', '/1/public/kafka/{}'.format(name), json=request)
        if resp.status_code == 200:
            return
        else:
            self._handle_error_response(resp)

    def kafka_tables(self):
        """List kafka tables.
        """
        resp = self._call_kafka('GET', '/1/public/kafka')
        if resp.status_code == 200:
            return resp.json()
        else:
            self._handle_error_response(resp)

    def kafka_poll(self, name):
        """Polls kafka table.
        """
        resp = self._call_kafka('GET', '/1/public/kafka/{}'.format(name))
        if resp.status_code == 200:
            return resp.json()
        else:
            self._handle_error_response(resp)

    def kafka_drop(self, name, force=False):
        """Drops a kafka table.
        """
        resp = self._call_kafka('DELETE', '/1/public/kafka/{}?force={}'.format(name, str(force).lower()))
        if resp.status_code == 204:
            return
        else:
            self._handle_error_response(resp)

    # ====================
    # storage
    # ====================

    def storage_quota(self):
        """Gets quota info from storage server

        Returns:
            dict: example::

                {'cachesSize': 0,
                 'materializedViewsSize': 0,
                 'tableIndexesSize': 0,
                 'tablesSize': 0,
                 'temporaryLocationsSize': 0,
                 'totalSize': 1000000000}

        """
        resp = self._call_storage('GET', '/1/public/quota_info')
        if resp.status_code == 200:
            return resp.json()
        else:
            self._handle_error_response(resp)


    def list_caches(self):
        """List caches created for this user"""
        resp = self._call_storage('GET', '/1/public/cache_entries')
        if resp.status_code == 200:
            return resp.json()
        else:
            self._handle_error_response(resp)

    def invalidate_caches(self):
        """Invalidates all caches"""
        resp = self._call_storage('POST', '/1/public/invalidate_caches')
        if resp.status_code == 204:
            return
        else:
            self._handle_error_response(resp)

    # ====================
    # Diagnostics
    # ====================

    def version(self):
        """Show version of executor server.
        """
        resp = self._call_executor('GET', '/version')
        if resp.status_code == 200:
            return resp.text.strip()
        else:
            self._handle_error_response(resp)

    def health(self):
        """Check if services are healthy.
        """
        self.executor_health()
        self.creds_health()

    def executor_health(self):
        """Check if executor service is healthy.
        """
        resp = self._call_executor('GET', '/health')
        if resp.status_code == 204:
            return
        else:
            self._handle_error_response(resp)

    def creds_health(self):
        """Check if credentials service is healthy.
        """
        resp = self._call_creds('GET', '/health')
        if resp.status_code == 204:
            return
        else:
            self._handle_error_response(resp)

    def storage_health(self):
        """Check if storage service is alive.
        """
        resp = self._call_storage('GET', '/health')
        if resp.status_code == 204:
            return
        else:
            self._handle_error_response(resp)

    def kafka_health(self):
        """Check if kafka service is alive.
        """
        resp = self._call_kafka('GET', '/health')
        if resp.status_code == 204:
            return
        else:
            self._handle_error_response(resp)

    # ====================
    # Query Logs
    # ====================
    def search_query_logs(self, limit=1000, start=None, end=None, state=None, keyword=None):
        """Lists query logs in the server
        Args:
            limit (int): If defined limits maximum number of results to return, defaults to 1000.
            start (datetime): If defined limits search results with only queries created after "start", defaults to None.
            end (datetime): If defined limits search results with only queries created before "end", defaults to None.
            state (str): If defined limits search results with only queries which finished in "state", defaults to None.
            keyword (str): If defined limits search results with only queries which contain "keyword".
        Returns:
            list: List with results
        """

        str_start = start.strftime("%Y-%m-%dT%H:%M:%S.%f") if start else None
        str_end = end.strftime("%Y-%m-%dT%H:%M:%S.%f") if end else None
        params = dict(n=limit, start=str_start, end=str_end, state=state, keyword=keyword)
        resp = self._call_executor('GET', '/1/public/query-logs', params=params)
        if resp.status_code == 200:
            return resp.json()
        else:
            self._handle_error_response(resp)

    def get_query_logs(self, query_id):
        """Gets query logs for a certain query-id

        Returns:
            dict: Dict with results, if query_id does not exist returns None.
        """
        resp = self._call_executor('GET', f'/1/public/query-logs/{query_id}/entries')
        if resp.status_code == 200:
            return resp.json()
        else:
            self._handle_error_response(resp)

interval_regex = re.compile('P(\\d+Y)?(\\d+M)?(\\d+W)?(\\d+D)?T?(\\d+H)?(\d+M)?(\\d+(\\.\\d+)?S)?')


def convert_types(data, tipe, np_arrays=False, use_relativedeltas=True):
    if data is None:
        return None
    elif tipe['type'] == 'timestamp':
        # 1975-06-23 23:00:00.234
        data = datetime.strptime(data, '%Y-%m-%dT%H:%M:%S.%f')
    elif tipe['type'] == 'date':
        data = datetime.strptime(data, '%Y-%m-%d').date()
    elif tipe['type'] == 'time':
        # we are missing the milliseconds for time, a bug in the server?
        # a problem with jackson serialization?
        data = datetime.strptime(data, '%H:%M:%S.%f').time()
    elif tipe['type'] == 'interval':
        # P1Y2M3DT4H5M6.007S
        m = interval_regex.match(data)
        def value(group):
            return float(m.group(group)[:-1]) if m.group(group) else 0

        if use_relativedeltas:
            # Creating a dateutil.relativedelta instead of a timedelta to support months and years
            data = relativedelta(years=value(1), months=value(2), weeks=value(3), days=value(4), hours=value(5),
                                 minutes=value(6), seconds=value(7))
        # Timedeltas do not have month or year
        elif value(1) or value(2):
            raise RawException(
                'Cannot create timedeltas with years or months. Consider the option use_relativedeltas=True')
        else:
            data = timedelta(weeks=value(3), days=value(4), hours=value(5), minutes=value(6), seconds=value(7))

    elif tipe['type'] == 'array':
        def recurse(l, n):
            if n == 0:
                return convert_types(l, tipe['cellType'], np_arrays, use_relativedeltas)
            else:
                return [recurse(l2, n - 1) for l2 in l]

        if np_arrays:
            import numpy as np
            data = np.array(recurse(data['data'], len(tipe['dimensions'])))
        else:
            data = recurse(data['data'], len(tipe['dimensions']))

    elif tipe['type'] == 'collection':
        if np_arrays:
            import numpy as np
            data = np.array([convert_types(d, tipe['inner'], np_arrays, use_relativedeltas) for d in data])
        else:
            data = [convert_types(d, tipe['inner'], np_arrays, use_relativedeltas) for d in data]
    elif tipe['type'] == 'record':
        for att in tipe['atts']:
            data[att['idn']] = convert_types(data[att['idn']], att['type'], np_arrays, use_relativedeltas)
    elif tipe['type'] == 'binary':
        data = base64.b64decode(data)
    return data
